/*
 * This file was generated automatically by ExtUtils::ParseXS version 3.43 from the
 * contents of SVD.xs. Do not edit this file, edit SVD.xs instead.
 *
 *    ANY CHANGES MADE HERE WILL BE LOST!
 *
 */

#line 1 "SVD.xs"
#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"

#include "ppport.h"

#include "const-c.inc"

#include <math.h>

//===================================================================
//
// Constants and Type Declarations
//
//===================================================================

// #define MAX_FEATURES    64            // Number of features to use 
#define MAX_FEATURES    8
// #define MIN_EPOCHS      120           // Minimum number of epochs per feature
#define MIN_EPOCHS      500
// #define MAX_EPOCHS      200           // Max epochs per feature
#define MAX_EPOCHS      2000

#define MIN_IMPROVEMENT 0.005        // Minimum improvement required to continue current feature
#define INITx           0.1           // Initialization value for features
#define LRATE           0.001         // Learning rate parameter
#define K               0.015         // Regularization parameter used to minimize over-fitting


//////////////////////////////////////////////////////////////////////////////////////////////////////

typedef unsigned char BYTE;
#define true 1
#define false 0

struct Movie
{
    int         RatingCount;
    int         RatingSum;
    double      RatingAvg;            
    double      PseudoAvg;
};
            // PseudoAvg is a weighted average used to deal with small movie counts 

struct Customer
{
    int         RatingCount;
    int         RatingSum;
};
    // int         CustomerId;

struct Data
{
    int         CustId;
    short       MovieId;
    BYTE        Rating;
    float       Cache;
};

// class Engine 

int num_customers;
int num_movies;
int num_ratings;

// private:

int             RatingCount;                                 // Current number of loaded ratings
struct Data     *Ratings;                                    // Array of ratings data
struct Movie    *Movies;                                     // Array of movie metrics
struct Customer *Customers;                                  // Array of customer metrics
float           **MovieFeatures;     // Array of features by movie (using floats to save space)
float           **CustFeatures;   // Array of features by customer (using floats to save space)

// was private, should be public now

void Engine(int, int, int);
void DestroyEngine();
void set_Movies(int, int, int);
void set_Ratings(int, int, int);
void CalcMetrics();
void CalcFeatures();
inline double PredictRating(short, int);
inline double PredictRating2(short, int, int, float, bool);

// end class Engine

// sdw accessors for Perl

void set_Movies(int movieId, int count, int sum) {
    Movies[movieId].RatingCount = count;
    Movies[movieId].RatingSum = sum;
}

void set_Ratings(int movieId, int custId, int rating) {
    Ratings[RatingCount].MovieId = (short)movieId;
    Ratings[RatingCount].CustId = custId;
    Ratings[RatingCount].Rating = (BYTE)rating;
    Ratings[RatingCount].Cache = 0;
    RatingCount++;
}

//===================================================================
//
// Engine Class 
//
//===================================================================

//-------------------------------------------------------------------
// Initialization
//-------------------------------------------------------------------

void Engine(int x_num_customers, int x_num_ratings, int x_num_movies) {

    int i, f;

    num_customers = x_num_customers;  // yeah, I know... there's a trick for doing this.  don't know it.  XXX todo.
    num_ratings = x_num_ratings;
    num_movies = x_num_movies; 

    RatingCount = 0;

    Ratings = (struct Data*)calloc(sizeof(struct Data), num_ratings);
    Movies = (struct Movie*)calloc(sizeof(struct Movie), num_movies);
    Customers = (struct Customer*)calloc(sizeof(struct Customer), num_customers);

    MovieFeatures = (float **)calloc(sizeof(float *), MAX_FEATURES);  // calloc should have these 0 initialized
    CustFeatures = (float **)calloc(sizeof(float *), MAX_FEATURES); 
    
    for (f=0; f<MAX_FEATURES; f++)
    {
        MovieFeatures[f] = (float *)calloc(sizeof(float), num_movies);
        CustFeatures[f] = (float *)calloc(sizeof(float), num_customers);
        for (i=0; i<num_movies; i++) {
            MovieFeatures[f][i] = (float)INITx;
        }
        for (i=0; i<num_customers; i++) {
            CustFeatures[f][i] = (float)INITx;
        }
    }
}

void DestroyEngine() {
    free(Ratings);
    free(Movies);
    free(Customers);
    free(MovieFeatures);
    free(CustFeatures);
    RatingCount = 0;
}

//-------------------------------------------------------------------
// Calculations - This Paragraph contains all of the relevant code
//-------------------------------------------------------------------

//
// CalcMetrics
// - Loop through the history and pre-calculate metrics used in the training 
//
void CalcMetrics()
{
    int i, cid;
    // IdItr itr;

    wprintf(L"\nCalculating intermediate metrics\n");

    // Process each row in the training set
    for (i=0; i<RatingCount; i++)
    {
        struct Data* rating = Ratings + i;

        // Increment movie stats
        Movies[rating->MovieId].RatingCount++;
        Movies[rating->MovieId].RatingSum += rating->Rating;

        // killed map translate stuff here -- sdw
        Customers[rating->CustId].RatingCount++;
        Customers[rating->CustId].RatingSum += rating->Rating;
    }

    // Do a follow-up loop to calc movie averages
    for (i=0; i<num_movies; i++)
    {
        struct Movie* movie = Movies+i;
        movie->RatingAvg = movie->RatingSum / (1.0 * movie->RatingCount);
        movie->PseudoAvg = (3.23 * 25 + movie->RatingSum) / (25.0 + movie->RatingCount);
    }
}

//
// CalcFeatures
// - Iteratively train each feature on the entire data set
// - Once sufficient progress has been made, move on
//
void CalcFeatures()
{
    int f, e, i, custId, cnt = 0;
    struct Data* rating;
    double err, p, sq, rmse_last, rmse = 2.0;
    short movieId;
    float cf, mf;

    for (f=0; f<MAX_FEATURES; f++)
    {
        wprintf(L"\n--- Calculating feature: %d ---\n", f);

        // Keep looping until you have passed a minimum number 
        // of epochs or have stopped making significant progress 
        for (e=0; (e < MIN_EPOCHS) || (rmse <= rmse_last - MIN_IMPROVEMENT); e++)
        {
            cnt++;
            sq = 0;
            rmse_last = rmse;

            for (i=0; i<RatingCount; i++)
            {
                rating = Ratings + i;
                movieId = rating->MovieId;
                custId = rating->CustId;

                // Predict rating and calc error
                p = PredictRating2(movieId, custId, f, rating->Cache, true);
            // wprintf(L"         movieId=%d custId=%d rating=%d predicted=%1.2f finalpredicted=%1.2f\n", movieId, custId, rating->Rating, p, PredictRating(movieId, custId));
                err = (1.0 * rating->Rating - p);
                sq += err*err;
                
                // Cache off old feature values
                cf = CustFeatures[f][custId];
                mf = MovieFeatures[f][movieId];

                // Cross-train the features
                CustFeatures[f][custId] += (float)(LRATE * (err * mf - K * cf));
                MovieFeatures[f][movieId] += (float)(LRATE * (err * cf - K * mf));
            }
            
            rmse = sqrt(sq/RatingCount);
                  
            // wprintf(L"     cnt='%d' rmse='%f' />\n",cnt,rmse);
        }

        // Cache off old predictions
        for (i=0; i<RatingCount; i++)
        {
            rating = Ratings + i;
            rating->Cache = (float)PredictRating2(rating->MovieId, rating->CustId, f, rating->Cache, false);
        }            
    }
}

//
// PredictRating
// - During training there is no need to loop through all of the features
// - Use a cache for the leading features and do a quick calculation for the trailing
// - The trailing can be optionally removed when calculating a new cache value
//
double PredictRating2(short movieId, int custId, int feature, float cache, bool bTrailing)
{
    // Get cached value for old features or default to an average
    double sum = (cache > 0) ? cache : 1; //Movies[movieId].PseudoAvg; 

    // Add contribution of current feature
    sum += MovieFeatures[feature][movieId] * CustFeatures[feature][custId];
    if (sum > 5) sum = 5;
    if (sum < 1) sum = 1;

    // Add up trailing defaults values
    if (bTrailing)
    // if (bTrailing && feature > 0)  // sdw -- no way to test this without more complicated data so leaving it alone for now
    {
        sum += (MAX_FEATURES-feature-1) * (INITx * INITx); // sdw -- this is bizarre; wrap around from the other side minus one?
        // sum += (feature-1) * (INITx * INITx); // sdw -- wouldn't this be more correct?  meh?
        if (sum > 5) sum = 5;
        if (sum < 1) sum = 1;
    }

    return sum;
}

//
// PredictRating
// - This version is used for calculating the final results
// - It loops through the entire list of finished features
//
double PredictRating(short movieId, int custId)
{
    double sum = 1; //Movies[movieId].PseudoAvg;
    int f;

    for (f=0; f<MAX_FEATURES; f++) 
    {
        sum += MovieFeatures[f][movieId] * CustFeatures[f][custId];
    }

    if (sum > 5) sum = 5;
    if (sum < 1) sum = 1;

    return sum;
}


/* 
###############################################################################
#
# SVD Sample Code
#
# Copyright (C) 2007 Timely Development (www.timelydevelopment.com)
#
# Special thanks to Simon Funk and others from the Netflix Prize contest 
# for providing pseudo-code and tuning hints.
#
# Feel free to use this code as you wish as long as you include 
# these notices and attribution. 
#
# Also, if you have alternative types of algorithms for accomplishing 
# the same goal and would like to contribute, please share them as well :)
#
# STANDARD DISCLAIMER:
#
# - THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY
# - OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT
# - LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND#OR
# - FITNESS FOR A PARTICULAR PURPOSE.
#
###############################################################################

Copyright  2008 by Timely Development, LLC.

 */

#line 340 "SVD.c"
#ifndef PERL_UNUSED_VAR
#  define PERL_UNUSED_VAR(var) if (0) var = var
#endif

#ifndef dVAR
#  define dVAR		dNOOP
#endif


/* This stuff is not part of the API! You have been warned. */
#ifndef PERL_VERSION_DECIMAL
#  define PERL_VERSION_DECIMAL(r,v,s) (r*1000000 + v*1000 + s)
#endif
#ifndef PERL_DECIMAL_VERSION
#  define PERL_DECIMAL_VERSION \
	  PERL_VERSION_DECIMAL(PERL_REVISION,PERL_VERSION,PERL_SUBVERSION)
#endif
#ifndef PERL_VERSION_GE
#  define PERL_VERSION_GE(r,v,s) \
	  (PERL_DECIMAL_VERSION >= PERL_VERSION_DECIMAL(r,v,s))
#endif
#ifndef PERL_VERSION_LE
#  define PERL_VERSION_LE(r,v,s) \
	  (PERL_DECIMAL_VERSION <= PERL_VERSION_DECIMAL(r,v,s))
#endif

/* XS_INTERNAL is the explicit static-linkage variant of the default
 * XS macro.
 *
 * XS_EXTERNAL is the same as XS_INTERNAL except it does not include
 * "STATIC", ie. it exports XSUB symbols. You probably don't want that
 * for anything but the BOOT XSUB.
 *
 * See XSUB.h in core!
 */


/* TODO: This might be compatible further back than 5.10.0. */
#if PERL_VERSION_GE(5, 10, 0) && PERL_VERSION_LE(5, 15, 1)
#  undef XS_EXTERNAL
#  undef XS_INTERNAL
#  if defined(__CYGWIN__) && defined(USE_DYNAMIC_LOADING)
#    define XS_EXTERNAL(name) __declspec(dllexport) XSPROTO(name)
#    define XS_INTERNAL(name) STATIC XSPROTO(name)
#  endif
#  if defined(__SYMBIAN32__)
#    define XS_EXTERNAL(name) EXPORT_C XSPROTO(name)
#    define XS_INTERNAL(name) EXPORT_C STATIC XSPROTO(name)
#  endif
#  ifndef XS_EXTERNAL
#    if defined(HASATTRIBUTE_UNUSED) && !defined(__cplusplus)
#      define XS_EXTERNAL(name) void name(pTHX_ CV* cv __attribute__unused__)
#      define XS_INTERNAL(name) STATIC void name(pTHX_ CV* cv __attribute__unused__)
#    else
#      ifdef __cplusplus
#        define XS_EXTERNAL(name) extern "C" XSPROTO(name)
#        define XS_INTERNAL(name) static XSPROTO(name)
#      else
#        define XS_EXTERNAL(name) XSPROTO(name)
#        define XS_INTERNAL(name) STATIC XSPROTO(name)
#      endif
#    endif
#  endif
#endif

/* perl >= 5.10.0 && perl <= 5.15.1 */


/* The XS_EXTERNAL macro is used for functions that must not be static
 * like the boot XSUB of a module. If perl didn't have an XS_EXTERNAL
 * macro defined, the best we can do is assume XS is the same.
 * Dito for XS_INTERNAL.
 */
#ifndef XS_EXTERNAL
#  define XS_EXTERNAL(name) XS(name)
#endif
#ifndef XS_INTERNAL
#  define XS_INTERNAL(name) XS(name)
#endif

/* Now, finally, after all this mess, we want an ExtUtils::ParseXS
 * internal macro that we're free to redefine for varying linkage due
 * to the EXPORT_XSUB_SYMBOLS XS keyword. This is internal, use
 * XS_EXTERNAL(name) or XS_INTERNAL(name) in your code if you need to!
 */

#undef XS_EUPXS
#if defined(PERL_EUPXS_ALWAYS_EXPORT)
#  define XS_EUPXS(name) XS_EXTERNAL(name)
#else
   /* default to internal */
#  define XS_EUPXS(name) XS_INTERNAL(name)
#endif

#ifndef PERL_ARGS_ASSERT_CROAK_XS_USAGE
#define PERL_ARGS_ASSERT_CROAK_XS_USAGE assert(cv); assert(params)

/* prototype to pass -Wmissing-prototypes */
STATIC void
S_croak_xs_usage(const CV *const cv, const char *const params);

STATIC void
S_croak_xs_usage(const CV *const cv, const char *const params)
{
    const GV *const gv = CvGV(cv);

    PERL_ARGS_ASSERT_CROAK_XS_USAGE;

    if (gv) {
        const char *const gvname = GvNAME(gv);
        const HV *const stash = GvSTASH(gv);
        const char *const hvname = stash ? HvNAME(stash) : NULL;

        if (hvname)
	    Perl_croak_nocontext("Usage: %s::%s(%s)", hvname, gvname, params);
        else
	    Perl_croak_nocontext("Usage: %s(%s)", gvname, params);
    } else {
        /* Pants. I don't think that it should be possible to get here. */
	Perl_croak_nocontext("Usage: CODE(0x%" UVxf ")(%s)", PTR2UV(cv), params);
    }
}
#undef  PERL_ARGS_ASSERT_CROAK_XS_USAGE

#define croak_xs_usage        S_croak_xs_usage

#endif

/* NOTE: the prototype of newXSproto() is different in versions of perls,
 * so we define a portable version of newXSproto()
 */
#ifdef newXS_flags
#define newXSproto_portable(name, c_impl, file, proto) newXS_flags(name, c_impl, file, proto, 0)
#else
#define newXSproto_portable(name, c_impl, file, proto) (PL_Sv=(SV*)newXS(name, c_impl, file), sv_setpv(PL_Sv, proto), (CV*)PL_Sv)
#endif /* !defined(newXS_flags) */

#if PERL_VERSION_LE(5, 21, 5)
#  define newXS_deffile(a,b) Perl_newXS(aTHX_ a,b,file)
#else
#  define newXS_deffile(a,b) Perl_newXS_deffile(aTHX_ a,b)
#endif

#line 484 "SVD.c"

/* INCLUDE:  Including 'const-xs.inc' from 'SVD.xs' */


XS_EUPXS(XS_Math__Preference__SVD_constant); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Math__Preference__SVD_constant)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "sv");
    PERL_UNUSED_VAR(ax); /* -Wall */
    SP -= items;
    {
#line 4 "./const-xs.inc"
#ifdef dXSTARG
	dXSTARG; /* Faster if we have it.  */
#else
	dTARGET;
#endif
	STRLEN		len;
        int		type;
	/* IV		iv;	Uncomment this if you need to return IVs */
	/* NV		nv;	Uncomment this if you need to return NVs */
	/* const char	*pv;	Uncomment this if you need to return PVs */
#line 509 "SVD.c"
	SV *	sv = ST(0)
;
	const char *	s = SvPV(sv, len);
#line 18 "./const-xs.inc"
	type = constant(aTHX_ s, len);
      /* Return 1 or 2 items. First is error message, or undef if no error.
           Second, if present, is found value */
        switch (type) {
        case PERL_constant_NOTFOUND:
          sv =
	    sv_2mortal(newSVpvf("%s is not a valid Math::Preference::SVD macro", s));
          PUSHs(sv);
          break;
        case PERL_constant_NOTDEF:
          sv = sv_2mortal(newSVpvf(
	    "Your vendor has not defined Math::Preference::SVD macro %s, used",
				   s));
          PUSHs(sv);
          break;
	/* Uncomment this if you need to return IVs
        case PERL_constant_ISIV:
          EXTEND(SP, 2);
          PUSHs(&PL_sv_undef);
          PUSHi(iv);
          break; */
	/* Uncomment this if you need to return NOs
        case PERL_constant_ISNO:
          EXTEND(SP, 2);
          PUSHs(&PL_sv_undef);
          PUSHs(&PL_sv_no);
          break; */
	/* Uncomment this if you need to return NVs
        case PERL_constant_ISNV:
          EXTEND(SP, 2);
          PUSHs(&PL_sv_undef);
          PUSHn(nv);
          break; */
	/* Uncomment this if you need to return PVs
        case PERL_constant_ISPV:
          EXTEND(SP, 2);
          PUSHs(&PL_sv_undef);
          PUSHp(pv, strlen(pv));
          break; */
	/* Uncomment this if you need to return PVNs
        case PERL_constant_ISPVN:
          EXTEND(SP, 2);
          PUSHs(&PL_sv_undef);
          PUSHp(pv, iv);
          break; */
	/* Uncomment this if you need to return SVs
        case PERL_constant_ISSV:
          EXTEND(SP, 2);
          PUSHs(&PL_sv_undef);
          PUSHs(sv);
          break; */
	/* Uncomment this if you need to return UNDEFs
        case PERL_constant_ISUNDEF:
          break; */
	/* Uncomment this if you need to return UVs
        case PERL_constant_ISUV:
          EXTEND(SP, 2);
          PUSHs(&PL_sv_undef);
          PUSHu((UV)iv);
          break; */
	/* Uncomment this if you need to return YESs
        case PERL_constant_ISYES:
          EXTEND(SP, 2);
          PUSHs(&PL_sv_undef);
          PUSHs(&PL_sv_yes);
          break; */
        default:
          sv = sv_2mortal(newSVpvf(
	    "Unexpected return type %d while processing Math::Preference::SVD macro %s, used",
               type, s));
          PUSHs(sv);
        }
#line 586 "SVD.c"
	PUTBACK;
	return;
    }
}


/* INCLUDE: Returning to 'SVD.xs' from 'const-xs.inc' */


XS_EUPXS(XS_Math__Preference__SVD_set_Movies); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Math__Preference__SVD_set_Movies)
{
    dVAR; dXSARGS;
    if (items != 3)
       croak_xs_usage(cv,  "movieId, count, sum");
    PERL_UNUSED_VAR(ax); /* -Wall */
    SP -= items;
    {
	int	movieId = (int)SvIV(ST(0))
;
	int	count = (int)SvIV(ST(1))
;
	int	sum = (int)SvIV(ST(2))
;
#line 343 "SVD.xs"
	I32* temp;
#line 613 "SVD.c"
#line 345 "SVD.xs"
	temp = PL_markstack_ptr++;
	set_Movies(movieId, count, sum);
	if (PL_markstack_ptr != temp) {
          /* truly void, because dXSARGS not invoked */
	  PL_markstack_ptr = temp;
	  XSRETURN_EMPTY; /* return empty stack */
        }
        /* must have used dXSARGS; list context implied */
	return; /* assume stack size is correct */
#line 624 "SVD.c"
	PUTBACK;
	return;
    }
}


XS_EUPXS(XS_Math__Preference__SVD_set_Ratings); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Math__Preference__SVD_set_Ratings)
{
    dVAR; dXSARGS;
    if (items != 3)
       croak_xs_usage(cv,  "movieId, custId, rating");
    PERL_UNUSED_VAR(ax); /* -Wall */
    SP -= items;
    {
	int	movieId = (int)SvIV(ST(0))
;
	int	custId = (int)SvIV(ST(1))
;
	int	rating = (int)SvIV(ST(2))
;
#line 361 "SVD.xs"
	I32* temp;
#line 648 "SVD.c"
#line 363 "SVD.xs"
	temp = PL_markstack_ptr++;
	set_Ratings(movieId, custId, rating);
	if (PL_markstack_ptr != temp) {
          /* truly void, because dXSARGS not invoked */
	  PL_markstack_ptr = temp;
	  XSRETURN_EMPTY; /* return empty stack */
        }
        /* must have used dXSARGS; list context implied */
	return; /* assume stack size is correct */
#line 659 "SVD.c"
	PUTBACK;
	return;
    }
}


XS_EUPXS(XS_Math__Preference__SVD_Engine); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Math__Preference__SVD_Engine)
{
    dVAR; dXSARGS;
    if (items != 3)
       croak_xs_usage(cv,  "x_num_customers, x_num_ratings, x_num_movies");
    PERL_UNUSED_VAR(ax); /* -Wall */
    SP -= items;
    {
	int	x_num_customers = (int)SvIV(ST(0))
;
	int	x_num_ratings = (int)SvIV(ST(1))
;
	int	x_num_movies = (int)SvIV(ST(2))
;
#line 379 "SVD.xs"
	I32* temp;
#line 683 "SVD.c"
#line 381 "SVD.xs"
	temp = PL_markstack_ptr++;
	Engine(x_num_customers, x_num_ratings, x_num_movies);
	if (PL_markstack_ptr != temp) {
          /* truly void, because dXSARGS not invoked */
	  PL_markstack_ptr = temp;
	  XSRETURN_EMPTY; /* return empty stack */
        }
        /* must have used dXSARGS; list context implied */
	return; /* assume stack size is correct */
#line 694 "SVD.c"
	PUTBACK;
	return;
    }
}


XS_EUPXS(XS_Math__Preference__SVD_DestroyEngine); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Math__Preference__SVD_DestroyEngine)
{
    dVAR; dXSARGS;
    if (items != 0)
       croak_xs_usage(cv,  "");
    PERL_UNUSED_VAR(ax); /* -Wall */
    SP -= items;
    {
#line 394 "SVD.xs"
	I32* temp;
#line 712 "SVD.c"
#line 396 "SVD.xs"
	temp = PL_markstack_ptr++;
	DestroyEngine();
	if (PL_markstack_ptr != temp) {
          /* truly void, because dXSARGS not invoked */
	  PL_markstack_ptr = temp;
	  XSRETURN_EMPTY; /* return empty stack */
        }
        /* must have used dXSARGS; list context implied */
	return; /* assume stack size is correct */
#line 723 "SVD.c"
	PUTBACK;
	return;
    }
}


XS_EUPXS(XS_Math__Preference__SVD_CalcMetrics); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Math__Preference__SVD_CalcMetrics)
{
    dVAR; dXSARGS;
    if (items != 0)
       croak_xs_usage(cv,  "");
    PERL_UNUSED_VAR(ax); /* -Wall */
    SP -= items;
    {
#line 409 "SVD.xs"
	I32* temp;
#line 741 "SVD.c"
#line 411 "SVD.xs"
	temp = PL_markstack_ptr++;
	CalcMetrics();
	if (PL_markstack_ptr != temp) {
          /* truly void, because dXSARGS not invoked */
	  PL_markstack_ptr = temp;
	  XSRETURN_EMPTY; /* return empty stack */
        }
        /* must have used dXSARGS; list context implied */
	return; /* assume stack size is correct */
#line 752 "SVD.c"
	PUTBACK;
	return;
    }
}


XS_EUPXS(XS_Math__Preference__SVD_CalcFeatures); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Math__Preference__SVD_CalcFeatures)
{
    dVAR; dXSARGS;
    if (items != 0)
       croak_xs_usage(cv,  "");
    PERL_UNUSED_VAR(ax); /* -Wall */
    SP -= items;
    {
#line 424 "SVD.xs"
	I32* temp;
#line 770 "SVD.c"
#line 426 "SVD.xs"
	temp = PL_markstack_ptr++;
	CalcFeatures();
	if (PL_markstack_ptr != temp) {
          /* truly void, because dXSARGS not invoked */
	  PL_markstack_ptr = temp;
	  XSRETURN_EMPTY; /* return empty stack */
        }
        /* must have used dXSARGS; list context implied */
	return; /* assume stack size is correct */
#line 781 "SVD.c"
	PUTBACK;
	return;
    }
}


XS_EUPXS(XS_Math__Preference__SVD_PredictRating2); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Math__Preference__SVD_PredictRating2)
{
    dVAR; dXSARGS;
    if (items != 5)
       croak_xs_usage(cv,  "movieId, custId, feature, cache, bTrailing");
    {
	short	movieId = (short)SvIV(ST(0))
;
	int	custId = (int)SvIV(ST(1))
;
	int	feature = (int)SvIV(ST(2))
;
	float	cache = (float)SvNV(ST(3))
;
	bool	bTrailing = (bool)SvTRUE(ST(4))
;
	double	RETVAL;
	dXSTARG;

	RETVAL = PredictRating2(movieId, custId, feature, cache, bTrailing);
	XSprePUSH; PUSHn((double)RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Math__Preference__SVD_PredictRating); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Math__Preference__SVD_PredictRating)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "movieId, custId");
    {
	short	movieId = (short)SvIV(ST(0))
;
	int	custId = (int)SvIV(ST(1))
;
	double	RETVAL;
	dXSTARG;

	RETVAL = PredictRating(movieId, custId);
	XSprePUSH; PUSHn((double)RETVAL);
    }
    XSRETURN(1);
}

#ifdef __cplusplus
extern "C"
#endif
XS_EXTERNAL(boot_Math__Preference__SVD); /* prototype to pass -Wmissing-prototypes */
XS_EXTERNAL(boot_Math__Preference__SVD)
{
#if PERL_VERSION_LE(5, 21, 5)
    dVAR; dXSARGS;
#else
    dVAR; dXSBOOTARGSXSAPIVERCHK;
#endif
#if PERL_VERSION_LE(5, 8, 999) /* PERL_VERSION_LT is 5.33+ */
    char* file = __FILE__;
#else
    const char* file = __FILE__;
#endif

    PERL_UNUSED_VAR(file);

    PERL_UNUSED_VAR(cv); /* -W */
    PERL_UNUSED_VAR(items); /* -W */
#if PERL_VERSION_LE(5, 21, 5)
    XS_VERSION_BOOTCHECK;
#  ifdef XS_APIVERSION_BOOTCHECK
    XS_APIVERSION_BOOTCHECK;
#  endif
#endif

        newXS_deffile("Math::Preference::SVD::constant", XS_Math__Preference__SVD_constant);
        newXS_deffile("Math::Preference::SVD::set_Movies", XS_Math__Preference__SVD_set_Movies);
        newXS_deffile("Math::Preference::SVD::set_Ratings", XS_Math__Preference__SVD_set_Ratings);
        newXS_deffile("Math::Preference::SVD::Engine", XS_Math__Preference__SVD_Engine);
        newXS_deffile("Math::Preference::SVD::DestroyEngine", XS_Math__Preference__SVD_DestroyEngine);
        newXS_deffile("Math::Preference::SVD::CalcMetrics", XS_Math__Preference__SVD_CalcMetrics);
        newXS_deffile("Math::Preference::SVD::CalcFeatures", XS_Math__Preference__SVD_CalcFeatures);
        newXS_deffile("Math::Preference::SVD::PredictRating2", XS_Math__Preference__SVD_PredictRating2);
        newXS_deffile("Math::Preference::SVD::PredictRating", XS_Math__Preference__SVD_PredictRating);
#if PERL_VERSION_LE(5, 21, 5)
#  if PERL_VERSION_GE(5, 9, 0)
    if (PL_unitcheckav)
        call_list(PL_scopestack_ix, PL_unitcheckav);
#  endif
    XSRETURN_YES;
#else
    Perl_xs_boot_epilog(aTHX_ ax);
#endif
}

